import pytest
import torch
import numpy as np
from numpy.random import randint
import random
from bisect import bisect_right
from torch.utils.data import (TensorDataset, ConcatDataset)

from audtorch.samplers import (
    buckets_of_even_size, buckets_by_boundaries, BucketSampler)
from audtorch.datasets.utils import defined_split


# data sets
data_size = 1000
num_feats = 160
max_length = 300
max_feature = 100
lengths = torch.randint(0, max_length, (data_size,))
inputs = torch.randint(0, max_feature, (data_size, num_feats, max_length))
data = TensorDataset(inputs)

# function params
num_buckets = randint(1, 10)
batch_sizes = num_buckets * [randint(1, np.iinfo(np.int8).max)]
num_batches = num_buckets * [randint(0, np.iinfo(np.int8).max)]
bucket_boundaries = [b + num for num, b in enumerate(
    sorted(list(random.sample(range(max_length - num_buckets),
                              (num_buckets - 1)))))]


@pytest.mark.parametrize("key_values", [lengths])
@pytest.mark.parametrize("num_buckets", [num_buckets])
@pytest.mark.parametrize("reverse", [True, False])
def test_buckets_of_even_size(key_values, num_buckets, reverse):

    expected_bucket_size, remainders = divmod(data_size, num_buckets)
    expected_bucket_dist = remainders * [expected_bucket_size + 1]
    expected_bucket_dist += (num_buckets - remainders) * [expected_bucket_size]

    buckets = buckets_of_even_size(key_values=key_values,
                                   num_buckets=num_buckets,
                                   reverse=reverse)

    expected_bucket_ids = set(range(num_buckets))
    bucket_ids = [buckets(i) for i in range(data_size)]
    key_values = [key_values[i] for i in range(data_size)]
    bucket_dist = [bucket_ids.count(bucket_id)
                   for bucket_id in range(num_buckets)]

    # do bucket ids only range from 0 to num_buckets-1?
    assert expected_bucket_ids == set(bucket_ids)

    sort_indices = sorted(range(data_size), key=lambda idx: key_values[idx],
                          reverse=reverse)
    sorted_buckets = [bucket_ids[idx] for idx in sort_indices]
    diff_buckets = np.diff(sorted_buckets)

    # sorted with monotonously increasing/decreasing length of key values?
    assert all(diff >= 0 for diff in diff_buckets)

    # are buckets evenly distributed (except for remainders)?
    assert expected_bucket_dist == bucket_dist


@pytest.mark.parametrize("key_values", [lengths])
@pytest.mark.parametrize("bucket_boundaries", [bucket_boundaries])
def test_buckets_by_boundaries(key_values, bucket_boundaries):

    buckets = buckets_by_boundaries(key_values=lengths,
                                    bucket_boundaries=bucket_boundaries)

    num_buckets = len(bucket_boundaries) + 1
    expected_bucket_ids = list(range(num_buckets))

    data_size = key_values.shape[0]
    bucket_ids = [buckets(idx) for idx in range(data_size)]
    key_values = [key_values[idx] for idx in range(data_size)]

    # do bucket ids only range from 0 to num_buckets-1?
    # missing ids only allowed if corresponding bucket empty
    missing_buckets = list(set(expected_bucket_ids) - set(bucket_ids))
    for bucket_id in missing_buckets:
        if bucket_id == 0:
            assert not any([v < bucket_boundaries[bucket_id]
                            for v in key_values])
        elif bucket_id == expected_bucket_ids[-1]:
            assert not any([v > bucket_boundaries[(bucket_id - 1)]
                            for v in key_values])
        else:
            assert not any([v in range(bucket_boundaries[(bucket_id - 1)],
                                       bucket_boundaries[bucket_id])
                            for v in key_values])

    sort_indices = sorted(range(data_size), key=lambda i: key_values[i])
    sorted_buckets = [bucket_ids[i] for i in sort_indices]
    diff_buckets = np.diff(sorted_buckets)

    # sorted with monotonously increasing/decreasing length of key values?
    assert all(diff >= 0 for diff in diff_buckets)


@pytest.mark.parametrize("data", [data])
@pytest.mark.parametrize(
    "key_func",
    [buckets_of_even_size(lengths, num_buckets, False),
     buckets_by_boundaries(lengths, bucket_boundaries)])
@pytest.mark.parametrize("expected_num_datasets", [num_buckets])
@pytest.mark.parametrize("batch_sizes", [batch_sizes])
@pytest.mark.parametrize("expected_num_batches", [num_batches])
@pytest.mark.parametrize("permuted_order", [False, random.shuffle(
    list(range(num_buckets)))])
@pytest.mark.parametrize("shuffle_each_bucket", [True, False])
@pytest.mark.parametrize("drop_last", [True, False])
def test_bucket_sampler(data, key_func, expected_num_datasets,
                        batch_sizes, expected_num_batches, permuted_order,
                        shuffle_each_bucket, drop_last):

    subsets = defined_split(data, key_func)
    concat_dataset = ConcatDataset(subsets)

    batch_sampler = BucketSampler(
        concat_dataset=concat_dataset,
        batch_sizes=batch_sizes,
        num_batches=expected_num_batches,
        permuted_order=permuted_order,
        shuffle_each_bucket=shuffle_each_bucket,
        drop_last=drop_last)

    num_datasets = len(batch_sampler.datasets)
    expected_dataset_ids = list(range(num_datasets))
    if isinstance(permuted_order, list) and permuted_order:
        expected_dataset_ids = permuted_order

    # assert data sets via batch sampler
    batch_indices = list(iter(batch_sampler))
    epoch_batch_sizes = [len(batch) for batch in batch_indices]
    dataset_sizes = [len(d) for d in batch_sampler.datasets]

    expected_epoch_batch_sizes = []
    expected_dset_ids = []

    for i in expected_dataset_ids:

        skip = False

        if batch_sizes[i] == 0 or expected_num_batches[i] == 0:
            continue

        fitted_batches = dataset_sizes[i] // batch_sizes[i]
        num_batches = fitted_batches
        if expected_num_batches[i] <= fitted_batches:
            num_batches = expected_num_batches[i]
            skip = True

        add_batch_sizes = [batch_sizes[i]] * num_batches

        if not drop_last and not skip:
            remainder = dataset_sizes[i] % batch_sizes[i]
            if remainder != 0:
                add_batch_sizes += [remainder]

        expected_epoch_batch_sizes += add_batch_sizes

        if len(add_batch_sizes) > 0:
            expected_dset_ids += [i]

    # are batch sizes as expected?
    assert expected_epoch_batch_sizes == epoch_batch_sizes

    unique_batch_ids = []
    for batch in batch_indices:
        ids = list(set([bisect_right(
            concat_dataset.cumulative_sizes, idx) for idx in batch]))
        unique_batch_ids += [ids]

    # are all samples of each batch from identical data set?
    assert all(list(map(lambda l: len(l) == 1, unique_batch_ids)))

    # flatten list
    unique_batch_ids = [i for ids in unique_batch_ids for i in ids]

    dataset_ids = []
    prev_id = None

    for i in unique_batch_ids:
        current_id = i
        if prev_id is None:
            dataset_ids += [current_id]
        else:
            if not current_id == prev_id:
                dataset_ids += [current_id]
        prev_id = current_id

    # are batches drawn from data sets in desired order?
    assert expected_dset_ids == dataset_ids



import os
from warnings import warn

import pandas as pd
import resampy
from tabulate import tabulate
from torch.utils.data import (Dataset, ConcatDataset)

from .utils import (ensure_same_sampling_rate, files_and_labels_from_df,
                    load, sampling_rate_after_transform, safe_path)


__doctest_skip__ = ['*']


class AudioDataset(Dataset):
    r"""Basic audio signal data set.

    This data set can be used if you have a list of files and a list of
    corresponding targets.

    In addition, this class is a great starting point to inherit from if you
    wish to build your own data set.

    * :attr:`transform` controls the input transform
    * :attr:`target_transform` controls the target transform
    * :attr:`files` controls the audio files of the data set
    * :attr:`targets` controls the corresponding targets
    * :attr:`sampling_rate` holds the sampling rate of the returned data
    * :attr:`original_sampling_rate` holds the sampling rate of the audio files
      of the data set

    Args:
        root (str): root directory of dataset.
        files (list): list of files
        targets (list): list of targets
        sampling_rate (int): sampling rate in Hz of the data set
        transform (callable, optional): function/transform applied on the
            signal. Default: `None`
        target_transform (callable, optional): function/transform applied on
            the target. Default: `None`

    Example:
        >>> data = AudioDataset(root='/data',
        ...                     files=['speech.wav', 'noise.wav'],
        ...                     targets=['speech', 'noise'],
        ...                     sampling_rate=8000)
        >>> print(data)
        Dataset AudioDataset
            Number of data points: 2
            Root Location: /data
            Sampling Rate: 8000Hz
        >>> signal, target = data[0]
        >>> target
        'speech'

    """

    def __init__(self, root, files, targets, sampling_rate, *, transform=None,
                 target_transform=None):
        self.root = safe_path(root)
        self.files = [os.path.join(self.root, f) for f in files]
        self.targets = targets
        self.original_sampling_rate = sampling_rate
        self.transform = transform
        self.target_transform = target_transform

        assert len(files) == len(targets)

        if not self._check_exists():
            raise RuntimeError('Dataset not found.')

    def __len__(self):
        return len(self.files)

    def __getitem__(self, index):
        signal, signal_sampling_rate = load(self.files[index])
        # Handle empty signals
        if signal.shape[1] == 0:
            warn('Returning previous file.', UserWarning)
            return self.__getitem__(index - 1)
        # Handle different sampling rate
        if signal_sampling_rate != self.original_sampling_rate:
            warn('Resample from {} to {}'
                 .format(signal_sampling_rate, self.original_sampling_rate),
                 UserWarning)
            signal = resampy.resample(signal, signal_sampling_rate,
                                      self.original_sampling_rate, axis=-1)

        target = self.targets[index]

        if self.transform is not None:
            signal = self.transform(signal)

        if self.target_transform is not None:
            target = self.target_transform(target)

        return signal, target

    @property
    def sampling_rate(self):
        return sampling_rate_after_transform(self)

    def _check_exists(self):
        return os.path.exists(self.root)

    def extra_repr(self):
        r"""Set the extra representation of the data set.

        To print customized extra information, you should reimplement
        this method in your own data set. Both single-line and multi-line
        strings are acceptable.

        The extra information will be shown after the sampling rate entry.

        """
        return ''

    def __repr__(self):
        fmt_str = 'Dataset ' + self.__class__.__name__ + '\n'
        fmt_str += '    Number of data points: {}\n'.format(self.__len__())
        fmt_str += '    Root Location: {}\n'.format(self.root)
        if self.sampling_rate == self.original_sampling_rate:
            fmt_str += '    Sampling Rate: {}Hz\n'.format(self.sampling_rate)
        else:
            fmt_str += ('    Sampling Rate: {}Hz (original: {}Hz)\n'
                        .format(self.sampling_rate,
                                self.original_sampling_rate))
        fmt_str += self.extra_repr()
        if self.transform:
            fmt_str += _include_repr('Transform', self.transform)
        if self.target_transform:
            fmt_str += _include_repr('Target Transform', self.target_transform)
        return fmt_str


class PandasDataset(AudioDataset):
    r"""Data set from pandas.DataFrame.

    Create a data set by accessing the file locations and corresponding labels
    through a pandas.DataFrame.

    You have to specify which labels of the data set you want as target by the
    names of the corresponding columns in the data frame. If you want to select
    one of those columns the label is returned directly in its corresponding
    data type or you can specify a list of columns and the data set will return
    a dictionary containing the labels.

    The filenames of the corresponding audio files have to be specified with
    relative path from `root` in the data frame in a column with the name
    :attr:`col_filename` which defaults to `filename`.

    * :attr:`transform` controls the input transform
    * :attr:`target_transform` controls the target transform
    * :attr:`files` controls the audio files of the data set
    * :attr:`targets` controls the corresponding targets
    * :attr:`sampling_rate` holds the sampling rate of the returned data
    * :attr:`original_sampling_rate` holds the sampling rate of the audio files
      of the data set
    * :attr:`column_labels` holds the name of the label columns

    Args:
        root (str): root directory of data set
        df (pandas.DataFrame): data frame with filenames and labels. Relative
            from `root`
        sampling_rate (int): sampling rate in Hz of the data set
        column_labels (str or list of str, optional): name of data frame
            column(s) containing the desired labels. Default: `label`
        column_filename (str, optional): name of column holding the file
            names. Default: `filename`
        transform (callable, optional): function/transform applied on the
            signal. Default: `None`
        target_transform (callable, optional): function/transform applied on
            the target. Default: `None`

    Example:
        >>> data = PandasDataset(root='/data',
        ...                      df=dataset_dataframe,
        ...                      sampling_rate=44100,
        ...                      column_labels='age')
        >>> print(data)
        Dataset AudioDataset
            Number of data points: 120
            Root Location: /data
            Sampling Rate: 44100Hz
            Label: age
        >>> signal, target = data[0]
        >>> target
        'age'

    """

    def __init__(self, root, df, sampling_rate, *, column_labels='label',
                 column_filename='filename', transform=None,
                 target_transform=None):
        files, labels = \
            files_and_labels_from_df(df, root=root,
                                     column_labels=column_labels,
                                     column_filename=column_filename)
        super().__init__(root, files, targets=labels,
                         sampling_rate=sampling_rate, transform=transform,
                         target_transform=target_transform)
        self.column_labels = column_labels

    def extra_repr(self):
        fmt_str = '    Labels: {}\n'.format(self.column_labels)
        return fmt_str


class CsvDataset(PandasDataset):
    r"""Data set from CSV files.

    Create a data set by reading the file locations and corresponding labels
    from a CSV file.

    You have to specify which labels you want as the target of the data set by
    the names of the corresponding columns in the CSV file. If you want to
    select one of those columns the target is returned directly in its
    corresponding data type or you can specify a list of columns and the data
    set will return a dictionary containing the targets.

    The filenames of the corresponding audio files have to be specified with
    relative path from `root` in the CSV file in a column with the name
    :attr:`col_filename` which defaults to `filename`.

    * :attr:`transform` controls the input transform
    * :attr:`target_transform` controls the target transform
    * :attr:`files` controls the audio files of the data set
    * :attr:`targets` controls the corresponding targets
    * :attr:`sampling_rate` holds the sampling rate of the returned data
    * :attr:`original_sampling_rate` holds the sampling rate of the audio files
      of the data set
    * :attr:`csv_file` holds the path to the used CSV file

    Args:
        root (str): root directory of data set
        csv_file (str): CSV file with filenames and labels. Relative from
            `root`
        sampling_rate (int): sampling rate in Hz of the data set
        column_labels (str or list of str, optional): name of CSV column(s)
            containing the desired labels. Default: `label`
        column_filename (str, optional): name of CSV column holding the file
            names. Default: `filename`
        sep (str, optional): CSV delimiter. Default: `,`
        transform (callable, optional): function/transform applied on the
            signal. Default: `None`
        target_transform (callable, optional): function/transform applied on
            the target. Default: `None`

    Example:
        >>> data = CsvDataset(root='/data',
        ...                   csv_file='train.csv',
        ...                   sampling_rate=44100,
        ...                   column_labels='age')
        >>> print(data)
        Dataset AudioDataset
            Number of data points: 120
            Root Location: /data
            Sampling Rate: 44100Hz
            Label: age
            CSV file: train.csv
        >>> signal, target = data[0]
        >>> target
        'age'

    """

    def __init__(self, root, csv_file, sampling_rate, *, sep=',',
                 column_labels='label', column_filename='filename',
                 transform=None, target_transform=None):
        self.root = safe_path(root)
        self.csv_file = os.path.join(self.root, csv_file)

        if not os.path.isfile(self.csv_file):
            raise FileNotFoundError('CSV file {} not found.'
                                    .format(self.csv_file))
        df = pd.read_csv(self.csv_file, sep)
        super().__init__(root, df, sampling_rate,
                         column_labels=column_labels,
                         column_filename=column_filename, transform=transform,
                         target_transform=target_transform)

    def extra_repr(self):
        fmt_str = super().extra_repr()
        fmt_str += ('    CSV file: {}\n'
                    .format(os.path.basename(self.csv_file)))
        return fmt_str


class AudioConcatDataset(ConcatDataset):
    r"""Concatenation data set of multiple audio data sets.

    This data set checks that all audio data sets are
    compatible with respect to the sampling rate which they
    are processed with.

    * :attr:`sampling_rate` holds the consistent sampling rate of the
      concatenated data set
    * :attr:`datasets` holds a list of all audio data sets
    * :attr:`cumulative_sizes` holds a list of sizes accumulated over all
      audio data sets, i.e. `[len(data1), len(data1) + len(data2), ...]`

    Args:
        datasets (list of audtorch.AudioDataset): Audio data sets
            with property `sampling_rate`.

    Example:
        >>> import sounddevice as sd
        >>> from audtorch.datasets import LibriSpeech
        >>> dev_clean = LibriSpeech(root='/data/LibriSpeech', sets='dev-clean')
        >>> dev_other = LibriSpeech(root='/data/LibriSpeech', sets='dev-other')
        >>> data = AudioConcatDataset([dev_clean, dev_other])
        >>> print(data)
        Data set AudioConcatDataset
        Number of data points: 5567
        Sampling Rate: 16000Hz
        <BLANKLINE>
        data sets      data points  extra
        -----------  -------------  ---------------
        LibriSpeech           2703  Sets: dev-clean
        LibriSpeech           2864  Sets: dev-other
        >>> signal, label = data[8]
        >>> label
        AS FOR ETCHINGS THEY ARE OF TWO KINDS BRITISH AND FOREIGN
        >>> sd.play(signal.transpose(), data.sampling_rate)

    """

    def __init__(self, datasets):
        super().__init__(datasets)
        ensure_same_sampling_rate(datasets)

    @property
    def sampling_rate(self):
        return self.datasets[0].sampling_rate

    def extra_repr(self):
        r"""Set the extra representation of the data set.

        To print customized extra information, you should reimplement
        this method in your own data set. Both single-line and multi-line
        strings are acceptable.

        The extra information will be shown after the sampling rate entry.

        """
        return ''

    def __repr__(self):
        fmt_str = 'Data set ' + self.__class__.__name__ + '\n'
        fmt_str += '    Number of data points: {}\n'.format(self.__len__())
        fmt_str += '    Sampling Rate: {}Hz\n'.format(self.sampling_rate)
        fmt_str += self.extra_repr()

        headers = ['data sets', 'data points',
                   'transform', 'target_transform', 'extra']
        tabular_data = [(d.__class__.__name__,
                         d.__len__(),
                         d.transform,
                         d.target_transform,
                         d.extra_repr()) for d in self.datasets]

        col = 0  # remove columns without any entries
        for _ in range(len(headers)):
            if not any([True if row[col] else False for row in tabular_data]):
                del headers[col]
                tabular_data = [row[:col] + row[col + 1:]
                                for row in tabular_data]
            else:
                col += 1

        fmt_str += '\n' + tabulate(tabular_data=tabular_data, headers=headers)
        return fmt_str


def _include_repr(name, obj):
    r"""Include __repr__ from other object as indented string.

    Args:
        name (str): Name of the object to be documented, e.g. "Transform".
        obj (object with `__repr__`): Object that provides `__repr__` output.

    Results:
        str: Format string of object to include into another `__repr__`.

    Example:
        >>> t = transforms.Pad(2)
        >>> datasets._include_repr('Transform', t)
        '    Transform: Pad(padding=2, value=0, axis=-1)\n'

    """
    part1 = '    {}: '.format(name)
    part2 = obj.__repr__().replace('\n', '\n' + ' ' * len(part1))
    return '{0}{1}\n'.format(part1, part2)



from copy import deepcopy
import threading
import queue

from tqdm import tqdm
import numpy as np


def flatten_list(nested_list):
    """Flatten an arbitrarily nested list.

    Implemented without  recursion to avoid stack overflows.
    Returns a new list, the original list is unchanged.

    Args:
        nested_list (list): nested list

    Returns:
        list: flattened list

    Example:
        >>> flatten_list([1, 2, 3, [4], [], [[[[[[[[[5]]]]]]]]]])
        [1, 2, 3, 4, 5]
        >>> flatten_list([[1, 2], 3])
        [1, 2, 3]

    """
    def _flat_generator(nested_list):
        while nested_list:
            sublist = nested_list.pop(0)
            if isinstance(sublist, list):
                nested_list = sublist + nested_list
            else:
                yield sublist
    nested_list = deepcopy(nested_list)
    return list(_flat_generator(nested_list))


def to_tuple(input, *, tuple_len=2):
    r"""Convert to tuple of given length.

    This utility function is used to convert single-value arguments to tuples
    of appropriate length, e.g. for multi-dimensional inputs where each
    dimension requires the same value. If the argument is already an iterable
    it is returned as a tuple if its length matches the desired tuple length.
    Otherwise a `ValueError` is raised.

    Args:
        input (non-iterable or iterable): argument to be converted to tuple
        tuple_len (int): required length of argument tuple. Default: `2`

    Returns:
        tuple: tuple of desired length

    Example:
        >>> to_tuple(2)
        (2, 2)

    """
    try:
        iter(input)
        if len(input) != tuple_len:
            raise ValueError('Input expected to be of length {} but was of '
                             'length {}'.format(tuple_len, len(input)))
        else:
            input = tuple(input)
    except TypeError:
        input = tuple([input] * tuple_len)
    return input


def energy(signal):
    r"""Energy of input signal.

    .. math::
        E = \sum_n |x_n|^2

    Args:
        signal (numpy.ndarray): signal

    Returns:
        float: energy of signal

    Example:
        >>> a = np.array([[2, 2]])
        >>> energy(a)
        8

    """
    return np.sum(np.abs(signal) ** 2)


def power(signal):
    r"""Power of input signal.

    .. math::
        P = {1 \over N} \sum_n |x_n|^2

    Args:
        signal (numpy.ndarray): signal

    Returns:
        float: power of signal

    Example:
        >>> a = np.array([[2, 2]])
        >>> power(a)
        4.0

    """
    return np.sum(np.abs(signal) ** 2) / signal.size


def run_worker_threads(num_workers, task_fun, params, *, progress_bar=False):
    r"""Run parallel tasks using worker threads.

    Args:
        num_workers (int): number of worker threads
        task_fun (Callable): task function with one or more
            parameters, e.g. x, y, z, and optionally returning a value
        params (list of tuples): list of tuples holding parameters
            for each task, e.g. [(x1, y1, z1), (x2, y2, z2), ...]
        progress_bar (bool): show a progress bar. Default: False

    Returns:
        list: result values in order of `params`

    Example:
        >>> power = lambda x, n: x ** n
        >>> params = [(2, n) for n in range(10)]
        >>> run_worker_threads(3, power, params)
        [1, 2, 4, 8, 16, 32, 64, 128, 256, 512]

    """
    num_workers = max(0, num_workers)
    n_tasks = len(params)
    results = [None] * n_tasks

    # do not use more workers as needed
    num_workers = n_tasks if num_workers == 0 else min(num_workers, n_tasks)

    # num_workers == 1 -> run sequentially
    if num_workers == 1:
        for index, param in enumerate(params):
            results[index] = task_fun(*param)

    # number_workers > 1 -> parallalize work
    else:

        # define worker thread
        def _worker():
            while True:
                item = q.get()
                if item is None:
                    break
                index, param = item
                results[index] = task_fun(*param)
                q.task_done()

        # create queue, possibly with a progress bar
        if progress_bar:
            class QueueWithProgbar(queue.Queue):
                def __init__(self, n_tasks, maxsize=0):
                    super().__init__(maxsize)
                    self.pbar = tqdm(total=n_tasks)

                def task_done(self):
                    super().task_done()
                    self.pbar.update(1)
            q = QueueWithProgbar(n_tasks)
        else:
            q = queue.Queue()

        # fill queue
        for index, param in enumerate(params):
            q.put((index, param))

        # start workers
        threads = []
        for i in range(num_workers):
            t = threading.Thread(target=_worker)
            t.start()
            threads.append(t)

        # block until all tasks are done
        q.join()

        # stop workers
        for _ in range(num_workers):
            q.put(None)
        for t in threads:
            t.join()

    return results



import numpy as np
try:
    import librosa
except ImportError:
    librosa = None

from .. import utils


def crop(signal, idx, *, axis=-1):
    r"""Crop signal along an axis.

    Args:
        signal (numpy.ndarray): audio signal
        idx (int or tuple): first (and last) index to return
        axis (int, optional): axis along to crop. Default: `-1`

    Note:
        Indexing from the end with `-1`, `-2`, ... is allowed. But you cannot
        use `-1` in the second part of the tuple to specify the last entry.
        Instead you have to write `(-2, signal.shape[axis])` to get the last
        two entries of `axis`, or simply `-1` if you only want to get the last
        entry.

    Returns:
        numpy.ndarray: cropped signal

    Example:
        >>> a = np.array([[1, 2], [3, 4]])
        >>> crop(a, 1)
        array([[2],
               [4]])

    """
    # Ensure idx is iterate able
    if isinstance(idx, int):
        idx = [idx]
    # Allow for -1 like syntax for index
    length = signal.shape[axis]
    idx = [length + i if i < 0 else i for i in idx]
    # Add stop index for single values
    if len(idx) == 1:
        idx = [idx[0], idx[0] + 1]

    # Split into three parts and return middle one
    return np.split(signal, idx, axis=axis)[1]


def pad(signal, padding, *, value=0, axis=-1):
    r"""Pad signal along an axis.

    If padding is an integer it pads equally on the left and right of the
    signal. If padding is a tuple with two entries it uses the first for the
    left side and the second for the right side.

    Args:
        signal (numpy.ndarray): audio signal
        padding (int or tuple): padding to apply on the left and right
        value (float, optional): value to pad with. Default: `0`
        axis (int, optional): axis along which to pad. Default: `-1`

    Returns:
        numpy.ndarray: padded signal

    Example:
        >>> a = np.array([[1, 2], [3, 4]])
        >>> pad(a, (0, 1))
        array([[1, 2, 0],
               [3, 4, 0]])

    """
    padding = utils.to_tuple(padding)
    dimensions = np.ndim(signal)
    pad = [(0, 0) for _ in range(dimensions)]  # no padding for all axes
    pad[axis] = padding                        # padding along selected axis

    return np.pad(signal, pad, 'constant', constant_values=value)


def replicate(signal, repetitions, *, axis=-1):
    r"""Replicate signal along an axis.

    Args:
        signal (numpy.ndarray): audio signal
        repetitions (int): number of times to replicate signal
        axis (int, optional): axis along which to replicate. Default: `-1`

    Returns:
        numpy.ndarray: replicated signal

    Example:
        >>> a = np.array([1, 2, 3])
        >>> replicate(a, 3)
        array([1, 2, 3, 1, 2, 3, 1, 2, 3])

    """
    reps = [1 for _ in range(len(signal.shape))]
    reps[axis] = repetitions
    return np.tile(signal, reps)


def downmix(signal, channels, *, method='mean', axis=-2):
    r"""Downmix signal to the provided number of channels.

    The downmix is done by one of these methods:

        * ``'mean'`` replace last desired channel by mean across itself and
          all remaining channels
        * ``'crop'`` drop all remaining channels

    Args:
        signal (numpy.ndarray): audio signal
        channels (int): number of desired channels
        method (str, optional): downmix method. Default: `'mean'`
        axis (int, optional): axis to downmix. Default: `-2`

    Returns:
        numpy.ndarray: reshaped signal

    Example:
        >>> a = np.array([[1, 2], [3, 4]])
        >>> downmix(a, 1)
        array([[2, 3]])

    """
    input_channels = np.atleast_2d(signal).shape[axis]

    if input_channels <= channels:
        return signal

    if method == 'mean':
        downmix = crop(signal, (channels - 1, input_channels), axis=axis)
        downmix = np.mean(downmix, axis=axis)
        signal = np.insert(signal, channels - 1, downmix, axis=axis)
    elif method == 'crop':
        pass
    else:
        raise TypeError('Method {} not supported.'.format(method))

    signal = crop(signal, (0, channels), axis=axis)
    return signal


def upmix(signal, channels, *, method='mean', axis=-2):
    r"""Upmix signal to the provided number of channels.

    The upmix is achieved by adding the same signal in the additional channels.
    The fixed signal is calculated by one of the following methods:

        * ``'mean'`` mean across all input channels
        * ``'zero'`` zeros
        * ``'repeat'`` last input channel

    Args:
        signal (numpy.ndarray): audio signal
        channels (int): number of desired channels
        method (str, optional): upmix method. Default: `'mean'`
        axis (int, optional): axis to upmix. Default: `-2`

    Returns:
        numpy.ndarray: reshaped signal

    Example:
        >>> a = np.array([[1, 2], [3, 4]])
        >>> upmix(a, 3)
        array([[1., 2.],
               [3., 4.],
               [2., 3.]])

    """
    input_channels = np.atleast_2d(signal).shape[axis]

    if input_channels >= channels:
        return signal

    signal = np.atleast_2d(signal)
    if method == 'mean':
        upmix = np.mean(signal, axis=axis)
        upmix = np.expand_dims(upmix, axis=axis)
    elif method == 'zero':
        upmix = np.zeros(signal.shape)
        upmix = crop(upmix, -1, axis=axis)
    elif method == 'repeat':
        upmix = crop(signal, -1, axis=axis)
    else:
        raise TypeError('Method {} not supported.'.format(method))

    upmix = np.repeat(upmix, channels - input_channels, axis=axis)
    return np.concatenate((signal, upmix), axis=axis)


def additive_mix(signal1, signal2, ratio):
    r"""Mix two signals additively by given ratio.

    If the power of one of the signals is below 1e-7, the signals are added
    without adjusting the signal-to-noise ratio.

    Args:
        signal1 (numpy.ndarray): audio signal
        signal2 (numpy.ndarray): audio signal
        ratio (int): ratio in dB of the second signal compared to the first one

    Returns:
        numpy.ndarray: mixture

    Example:
        >>> a = np.array([[1, 2], [3, 4]])
        >>> additive_mix(a, a, -10 * np.log10(0.5 ** 2))
        array([[1.5, 3. ],
               [4.5, 6. ]])

    """
    if signal1.shape != signal2.shape:
        raise ValueError('Shape of signal1 ({}) and signal2 ({}) do not'
                         ' match'.format(signal1.shape, signal2.shape))
    # If one of the signals includes only silence, don't apply SNR
    tol = 1e-7
    if utils.power(signal1) < tol or utils.power(signal2) < tol:
        scaling_factor = 1
    else:
        scaling_factor = (utils.power(signal1)
                          / utils.power(signal2)
                          * 10 ** (-ratio / 10))
    return signal1 + np.sqrt(scaling_factor) * signal2


def normalize(signal, *, axis=None):
    r"""Normalize signal.

    Ensure the maximum of the absolute value of the signal is 1.

    Note:
        The signal will never be divided by a number smaller than 1e-7.
        Meaning signals which are nearly silent are only slightly
        amplified.

    Args:
        signal (numpy.ndarray): audio signal
        axis (int, optional): normalize only along the given axis.
            Default: `None`

    Returns:
        numpy.ndarray: normalized signal

    Example:
        >>> a = np.array([[1, 2], [3, 4]])
        >>> normalize(a)
        array([[0.25, 0.5 ],
               [0.75, 1.  ]])

    """
    if axis is not None:
        peak = np.expand_dims(np.amax(np.abs(signal), axis=axis), axis=axis)
    else:
        peak = np.amax(np.abs(signal))
    return signal / np.maximum(peak, 1e-7)


def standardize(signal, *, mean=True, std=True, axis=None):
    r"""Standardize signal.

    Ensure the signal has a mean value of 0 and a variance of 1.

    Note:
        The signal will never be divided by a variance smaller than 1e-7.

    Args:
        signal (numpy.ndarray): audio signal
        mean (bool, optional): apply mean centering. Default: `True`
        std (bool, optional): normalize by standard deviation. Default: `True`
        axis (int, optional): standardize only along the given axis.
            Default: `None`

    Returns:
        numpy.ndarray: standardized signal

    Example:
        >>> a = np.array([[1, 2], [3, 4]])
        >>> standardize(a)
        array([[-1.34164079, -0.4472136 ],
               [ 0.4472136 ,  1.34164079]])

    """
    if mean:
        signal_mean = np.mean(signal, axis=axis)
        if axis is not None:
            signal_mean = np.expand_dims(signal_mean, axis=axis)
        signal = signal - signal_mean
    if std:
        signal_std = np.std(signal, axis=axis)
        if axis is not None:
            signal_std = np.expand_dims(signal_std, axis=axis)
        signal = signal / np.maximum(signal_std, 1e-7)
    return signal


def stft(signal, window_size, hop_size, *, fft_size=None, window='hann',
         axis=-1):
    r"""Short-time Fourier transform.

    The Short-time Fourier transform (STFT) is calculated by using librosa.
    It returns an array with the same shape as the input array, besides the
    axis chosen for STFT calculation is replaced by the two new ones of the
    spectrogram.

    The chosen FFT size is set identical to `window_size`.

    Args:
        signal (numpy.ndarray): audio signal
        window_size (int): size of STFT window in samples
        hop_size (int): size of STFT window hop in samples
        window (str, tuple, number, function, or numpy.ndarray, optional): type
            of STFT window. Default: `hann`
        axis (int, optional): axis of STFT calculation. Default: `-1`

    Returns:
        numpy.ndarray: complex spectrogram with the shape of its last two
        dimensions as `(window_size/2 + 1,
        np.ceil((len(signal) + window_size/2) / hop_size))`

    Example:
        >>> a = np.array([1., 2., 3., 4.])
        >>> stft(a, 2, 1)
        array([[ 1.+0.j,  2.+0.j,  3.+0.j,  4.+0.j,  3.+0.j],
               [-1.+0.j, -2.+0.j, -3.+0.j, -4.+0.j, -3.+0.j]], dtype=complex64)

    """
    samples = signal.shape[axis]
    if samples < window_size:
        raise ValueError('`signal` of length {} needs to be at least '
                         'as long as the `window_size` of {}'
                         .format(samples, window_size))

    # Pad to ensure same signal length after reconstruction
    # See discussion at https://github.com/librosa/librosa/issues/328
    signal = pad(signal, (0, np.mod(samples, hop_size)), value=0, axis=axis)
    if fft_size is None:
        fft_size = window_size
    fft_config = dict(n_fft=fft_size, hop_length=hop_size,
                      win_length=window_size, window=window)
    spectrogram = np.apply_along_axis(librosa.stft, axis, signal, **fft_config)
    return spectrogram


def istft(spectrogram, window_size, hop_size, *, window='hann', axis=-2):
    r"""Inverse Short-time Fourier transform.

    The inverse Short-time Fourier transform (iSTFT) is calculated by using
    librosa.
    It handles multi-dimensional inputs, but assumes that the two spectrogram
    axis are beside each other, starting with the axis corresponding to
    frequency bins.
    The returned audio signal has one dimension less than the spectrogram.

    Args:
        spectrogram (numpy.ndarray): complex spectrogram
        window_size (int): size of STFT window in samples
        hop_size (int): size of STFT window hop in samples
        window (str, tuple, number, function, or numpy.ndarray, optional): type
            of STFT window. Default: `hann`
        axis (int, optional): axis of frequency bins of the spectrogram. Time
            bins are expected at `axis + 1`. Default: `-2`

    Returns:
        numpy.ndarray: signal with shape `(number_of_time_bins * hop_size -
        window_size/2)`

    Example:
        >>> a = np.array([1., 2., 3., 4.])
        >>> D = stft(a, 4, 1)
        >>> istft(D, 4, 1)
        array([1., 2., 3., 4.], dtype=float32)

    """
    if axis == -1:
        raise ValueError('`axis` of spectrogram frequency bins cannot be -1')

    ifft_config = dict(hop_length=hop_size, win_length=window_size,
                       window=window)
    # Size of frequency and time axis
    f = spectrogram.shape[axis]
    t = spectrogram.shape[axis + 1]
    # Reshape the two axes of spectrogram into one axis
    shape_before = spectrogram.shape[:axis]
    shape_after = spectrogram.shape[axis:][2:]
    D = np.reshape(spectrogram, [*shape_before, f * t, *shape_after])
    # Adjust negative axis values as the second spectrogram axis was removed
    if axis < -1:
        axis += 1
    # iSTFT along the axis
    signal = np.apply_along_axis(_istft, axis, D, f, t, **ifft_config)
    # Remove padding that was added for STFT
    samples = signal.shape[axis]
    signal = crop(signal, (0, samples - np.mod(samples, hop_size)), axis=axis)
    return signal


def _istft(spectrogram, frequency_bins, time_bins, **config):
    """Inverse Short-time Fourier transform from a single axis.

    Time and frequency bins have to be provided in a single vector. This allows
    effective computation using `numpy.apply_along_axis`.

    Args:
        spectrogram (numpy.array): one dimensional vector
        frequency_bins (int): number of frequency bins
        time_bins (int): number of time bins
        **config (dict, optional): keyword arguments for librosa.istft

    Returns:
        numpy.array: time-series

    """
    # Reshape to [frequency_bins, time_bins] as expected by librosa
    spectrogram = np.reshape(spectrogram, [frequency_bins, time_bins])
    return librosa.istft(spectrogram, **config)
